///////////////////////////////////////////////////////////////////////////////
//
//  File Name 
//      MAIN.CPP
//
//  Description
//      This file implements the entry point for the DIRMON sample. It
//      also implements the initialization and cleanup of the MAPI session
//      we run in.
//
//  Author
//      Irving De la Cruz (irvingd@microsoft.com)
//
//  Revision: 1.1
//
// Written for Microsoft Corporation.
// Copyright (c) 1996-1997 Microsoft Corporation. All rights reserved.
//
#include "DIRMON.H"
#include <WINDOWSX.H>

LPMAPISESSION           g_pSession = NULL;
LPMDB                   g_pMDB = NULL;
LPMAPIFOLDER            g_pFolder = NULL;
SBinary                 g_IdentityEID = { 0 };
HINSTANCE               g_hInstance = NULL;
HANDLE                  g_hTerminateEvent = NULL;
SERVICE_STATUS_HANDLE   g_ssHandle = NULL;
DIRMON_CONFIG           g_Config = { 0 };

static char rgszCmdArgHelp[] =
{
    " Command Line Options                   Action\r\n"
    " /HELP or /?                   \tThis message box\r\n"
    " /APP                          \tRun this service as an normal application\r\n"
    " /REMOVE                       \tRemoves the service\r\n"
    " /INSTALL                      \tInstall the service\r\n"
    "\r\n"
    "     Parameters for the /INSTALL option\r\n"
    " /ACCOUNT              \tDomain account for the service\r\n"
    "                     \t\tin the form DOMAIN\\USER\r\n"
    " /PASSWORD             \tPassword for the account\r\n"
    " /PROFILE            \t\tMAPI profile the service should use\r\n"
    " /DIRECTORY            \tDirectory the service will monitor\r\n"
    "\r\n"
    " Example: (all in one line)\r\n"
    "\r\n"
    "   DIRMON /INSTALL /ACCOUNT=PDC\\IrvingD /PASSWORD=password\r\n"
    "           \t    /PROFILE=\"My Service Profile\" /DIRECTORY=C:\\TEMP\r\n"
};

#define MSGBOX_TITLE    "Directory Monitor Sample Service"

///////////////////////////////////////////////////////////////////////////////
//    GetCmdLineOptionValue()
//
//    Parameters
//      [IN, OUT]   pszCmdStart     String where the option starts. This
//                                  string should begin somewhere before
//                                  the "=" sign. If the VALUE is found, then
//                                  this string will have a NULL terminator
//                                  added after the VALUE substring.
//      
//    Purpose
//      Parses a command line option in the format /OPTION=VALUE and returns a
//      string pointer at the beginning of "VALUE".
//      
//    Return Value
//      The string with the command line value. NULL if nothing after the
//      "=" sign and before the next command line option.
//      
LPSTR WINAPI GetCmdLineOptionValue (LPSTR pszCmdStart)
{
    // Get rid of the /xxxx= prefixes and get to the actual value of the options
    while (*pszCmdStart && '=' != *pszCmdStart)
    {
        pszCmdStart++;
    }
    while (*pszCmdStart && ('=' == *pszCmdStart || ' ' == *pszCmdStart))
    {
        pszCmdStart++;
    }
    // For quoted values in the form /OPTION="Some Literal Value String"
    if (*pszCmdStart && '"' == *pszCmdStart)
    {
        pszCmdStart++;
        LPSTR pszQuotedValue = pszCmdStart;
        while (*pszCmdStart && '"' != *pszCmdStart)
        {
            pszCmdStart++;
        }
        if (*pszCmdStart && '"' == *pszCmdStart)
        {
            *pszCmdStart = 0;
            return pszQuotedValue;
        }
        // Un-ended quote
        return NULL;
    }
    return strtok (pszCmdStart, " =/");
}

///////////////////////////////////////////////////////////////////////////////
//    WinMain()
//
//    Parameters
//      { Refer to Win32 SDK documentation }
//      
//    Purpose
//      This is the entry function. The application will start executing here.
//      
//    Return Value
//      Return code of the application to the system.
//      
int WINAPI WinMain (HINSTANCE   hInst,
                    HINSTANCE   hPrevInst,
                    LPSTR       pszCmdLine,
                    int         nCmdLine)
{
    char szCmdLine[256] = { 0 };
    g_hInstance = hInst;

    // Before comparing the command line, change the string to UPPER case,
    // but before save the original command line because we want to preserve
    // the original case of the option values.
    lstrcpy (szCmdLine, pszCmdLine);
    CharUpper (pszCmdLine);

    // Handle the /HELP, /?, and ? options
    if (!lstrlen (pszCmdLine) ||
		strstr (pszCmdLine, CMDLINE_HELP) ||
        strstr (pszCmdLine, CMDLINE_HELP_ALT_1) ||
        strstr (pszCmdLine, CMDLINE_HELP_ALT_2))
    {
        MessageBox (NULL,
                    rgszCmdArgHelp,
                    MSGBOX_TITLE,
                    MB_OK | MB_ICONINFORMATION);
        return 0;
    }

    // Handle the /REMOVE option
    if (strstr (pszCmdLine, CMDLINE_REMOVE))
    {
        RemoveEventLogSource();
        RemoveService (DIRMON_SERVICE_NAME_SHORT);
        MessageBox (NULL,
                    "The service has been removed from the Service Control Manager list.",
                    MSGBOX_TITLE,
                    MB_OK | MB_ICONEXCLAMATION);
        return 0;
    }

    // Handle the /INSTALL ... option
    if (strstr (pszCmdLine, CMDLINE_INSTALL))
    {
        LPSTR pszAcct = strstr (pszCmdLine, CMDLINE_INSTALL_ACCOUNT);
        LPSTR pszPass = strstr (pszCmdLine, CMDLINE_INSTALL_PASSWORD);
        // This must be a profile name with no spaces:
        //      Invalid profile names: "Service Account", "Dirmon Profile"
        //      Valid profile names: "Service_Account", "Dirmon_Profile"
        LPSTR pszProf = strstr (pszCmdLine, CMDLINE_INSTALL_PROFILE);
        LPSTR pszDir  = strstr (pszCmdLine, CMDLINE_INSTALL_DIRECTORY);
        if (pszAcct && pszPass && pszProf && pszDir)
        {
            pszAcct = GetCmdLineOptionValue (pszAcct);
            // Before retriving the password, we must use the original
            // command line arguments because we want to preserve the
            // case of the /PASSWORD option
            DWORD i = pszPass - pszCmdLine;
            pszPass = GetCmdLineOptionValue (&szCmdLine[i]);
            i = pszProf - pszCmdLine;
            pszProf = GetCmdLineOptionValue (&szCmdLine[i]);
            i = pszDir - pszCmdLine;
            pszDir = GetCmdLineOptionValue (&szCmdLine[i]);
        }
         if (!pszAcct || !pszPass || !pszProf || !pszDir)
        {
            MessageBox (NULL,
                        "Missing argument for Command Line Option",
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONEXCLAMATION);
            MessageBox (NULL,
                        rgszCmdArgHelp,
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONINFORMATION);
            return ERROR_INVALID_PARAMETER;
        }

        char szDomainAndAccount[256] = { 0 };
        lstrcpy (szDomainAndAccount, pszAcct);
        LPSTR pszDomainOnly = strtok (pszAcct, "\\");
        pszAcct = strtok (NULL, "\\");
        if (!pszAcct)
        {
            MessageBox (NULL,
                        "The domain and account are in the wrong format. It should be DOMAIN\\USER.",
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONSTOP);
            MessageBox (NULL,
                        rgszCmdArgHelp,
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONINFORMATION);
            return ERROR_INVALID_PARAMETER;
        }
        char szFullPathEXE[MAX_PATH];
        GetModuleFileName (NULL, szFullPathEXE, MAX_PATH - 1);
        DWORD dwError = AddEventLogSource (szFullPathEXE);
        if (!dwError)
        {
            SetRequiredPrivileges (pszAcct);
            AddUserToLocalGroup (pszAcct, "Administrators");
            lstrcat (szFullPathEXE, " ");
            lstrcat (szFullPathEXE, CMDLINE_RUN_AS_SERVICE);
            dwError = InstallService (szFullPathEXE,
                                      DIRMON_SERVICE_NAME,
                                      DIRMON_SERVICE_NAME_SHORT,
                                      szDomainAndAccount,
                                      pszPass);
            if (!dwError)
            {
                dwError = AddConfigRegistryValues (pszProf, pszDir);
                // If there is an error adding the configuration
                // registry key, remove the service and bail out
                if (dwError)
                {
                    RemoveService (DIRMON_SERVICE_NAME_SHORT);
                }
            }
            // If there is an error installing the service,
            // remove the event log source
            if (dwError)
            {
                RemoveEventLogSource();
            }
        }
        if (dwError)
        {
            char szMsg[128];
            wsprintf (szMsg, "There was an error installing the service: Error code %d", dwError);
            MessageBox (NULL,
                        szMsg,
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONEXCLAMATION);
        }
        else
        {
            MessageBox (NULL,
                        "Service has been successfully installed",
                        MSGBOX_TITLE,
                        MB_OK | MB_ICONEXCLAMATION);
        }
        return dwError;
    }

    BOOL fRunAsService = FALSE;
    // Handle the /RUN_AS_SERVICE option
    if (strstr (pszCmdLine, CMDLINE_RUN_AS_SERVICE))
    {
        fRunAsService = TRUE;
    }

    // Is the application running as a SERVICE? If So, register the
    // entry point with the control dispatcher and get out.
    if (fRunAsService)
    {
        SERVICE_TABLE_ENTRY ServiceTable[] =
        {
            {
                DIRMON_SERVICE_NAME_SHORT,
                (LPSERVICE_MAIN_FUNCTION)DirMonServiceMain,
            },
            { NULL, NULL }
        };

        StartServiceCtrlDispatcher (ServiceTable);
        fRunAsService = TRUE;
    }
    else
    {
        // Looks like we are running as a Windows application, create
        // the "end" event and start the logon process (this occurs in
        // the WM_INITDIALOG message of the dialog procedure)
        g_hTerminateEvent = CreateEvent (0, TRUE, FALSE, 0);
        if (g_hTerminateEvent)
        {
            DialogBox (g_hInstance,
                       MAKEINTRESOURCE(IDD_DIRMON_DLG),
                       NULL,
                       ServiceDlgProc);
        }
    }
    // If we were running as a normal Windows application, cleanup before leaving
    if (!fRunAsService)
    {
        TerminateService (0, FALSE);
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
//    LoadConfigSettings()
//
//    Parameters
//      None
//
//    Purpose
//      Opens the registry key for this service and reads the current
//      settings into the global configuration structure
//
//    Return Value
//      A Win32 error code
//
DWORD WINAPI LoadConfigSettings()
{
    HKEY hKey;
    DWORD dwError = RegOpenKeyEx (HKEY_LOCAL_MACHINE,
                                  REGKEY_DIRMON_SERVICE_CONFIG,
                                  0,
                                  KEY_ALL_ACCESS,
                                  &hKey);
    if (dwError)
    {
        LogEvent (TRUE, EVENTMSG_E_OPEN_REG_CONFIG_KEY, dwError);
    }
    else
    {
        DWORD cbSize = sizeof (g_Config.szMAPIProfile);
        RegQueryValueEx (hKey,
                         REGKEY_CONFIG_PROFILE,
                         NULL,
                         NULL,
                         (LPBYTE)g_Config.szMAPIProfile,
                         &cbSize);
        cbSize = sizeof (g_Config.szMonitorDir);
        RegQueryValueEx (hKey,
                         REGKEY_CONFIG_MONDIR,
                         NULL,
                         NULL,
                         (LPBYTE)g_Config.szMonitorDir,
                         &cbSize);
        RegCloseKey (hKey);
    }
    return dwError;
}

///////////////////////////////////////////////////////////////////////////////
//    InitializeService()
//
//    Parameters
//      [IN]    hListBox    Handle to a list box where we display progress
//                          info. This handle is NULL if we are running in
//                          the SCM (Service Control Manager) context.
//
//    Purpose
//      Initializes the service threads and loads the configuration
//      settings from the registry
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI InitializeService (HWND hListBox)
{
    InitEventLog();
    if (hListBox)
    {
        ListBox_AddString (hListBox, "Loading registry settings...");
    }
    // Load the settings from the registry
    LoadConfigSettings();

    if (hListBox)
    {
        ListBox_AddString (hListBox, "Initializing MAPI session...");
    }
    HRESULT hResult = InitMAPISession ((hListBox ? FALSE : TRUE), g_Config.szMAPIProfile);
    if (hListBox)
    {
        ListBox_AddString (hListBox, "Session Initialized");
    }
    // If we are successful logging into the MAPI session, open the Exchange public store
    if (!hResult)
    {
        if (hListBox)
        {
            ListBox_AddString (hListBox, "Opening the public message store...");
        }
        hResult = OpenExchangePFStore();
        if (hListBox)
        {
            ListBox_AddString (hListBox, "Public message store opened");
        }
        // Now that the store is open, dispatch the worker thread 
        // will set the directory watching
        if (!hResult)
        {
            DWORD dwThreadID;
            HANDLE hThread = CreateThread (NULL, 
                                           0, 
                                           (LPTHREAD_START_ROUTINE)StartDirectoryWatch, 
                                           NULL,
                                           0, 
                                           &dwThreadID);
            CloseHandle (hThread);
        }
    }
    // If the handle to the listbox is valid, it means we are running
    // as an interactive app, not as a service. When this app is not
    // running as a service, this function runs on its own thread
    // which should block until the service is finish. The reason for
    // this is that the session clean up must be done on the same thread
    // that initialized it.
    if (hListBox)
    {
        ServiceWaitOnEvent (g_hTerminateEvent);
        FinishMAPISession();
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    AboutDlgProc()
//
//    Parameters
//      { Refer to Win32 API documentation on dialog procedures }
//
//    Purpose
//      Handle the UI of the ABOUT dialog box
//      
//    Return Value
//      TRUE if message was handled, FALSE if we don't handle the message
//      
BOOL CALLBACK AboutDlgProc (HWND   hDlg, 
                            UINT   message, 
                            WPARAM wParam, 
                            LPARAM lParam)
{
    if (WM_COMMAND == message)
    {
        if (IDCANCEL == LOWORD(wParam) || IDOK == LOWORD(wParam))
        {
            EndDialog (hDlg, TRUE);
        }
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    ServiceDlgProc()
//
//    Parameters
//      { Refer to Win32 API documentation on dialog procedures }
//
//    Purpose
//      Handle the message for the dialog of the service.
//      
//    Return Value
//      TRUE if message was handled, FALSE if we don't handle the message
//      
BOOL CALLBACK ServiceDlgProc (HWND    hDlg, 
                              UINT    message, 
                              WPARAM  wParam, 
                              LPARAM  lParam)
{
    static HANDLE hThread;
    DWORD dwThreadID;
    switch (message)
    {
        case WM_INITDIALOG :
            hThread = CreateThread (NULL, 
                                    0, 
                                    (LPTHREAD_START_ROUTINE)InitializeService, 
                                    GetDlgItem (hDlg, IDC_LISTBOX), 
                                    0, 
                                    &dwThreadID);
            break;

        case WM_COMMAND :
            switch (LOWORD(wParam))
            {
                case IDM_ABOUT :
                    DialogBox (g_hInstance,
                               MAKEINTRESOURCE(IDD_ABOUT),
                               NULL,
                               AboutDlgProc);
                    break;
                case IDM_FILE_EXIT :
                    PostMessage (hDlg, WM_SYSCOMMAND, SC_CLOSE, 0);
                    return TRUE;
            }
            break;

        case WM_SYSCOMMAND:
            if (wParam == SC_CLOSE)
            {
                FinishDirectoryWatch();
                SetEvent (g_hTerminateEvent);
                WaitForSingleObject (hThread, INFINITE);
                EndDialog (hDlg, TRUE);
                return TRUE;
            }
            break;
    }
    return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
//    InitMAPISession()
//
//    Parameters
//      [IN]    fIsNTService    TRUE if the application is running as an NT
//                              service. FALSE otherwise.
//      [IN]    pszProfileName  The name of the profile the service will
//                              logon. If this is NULL, the service tries
//                              to login using the default profile.
//
//    Purpose
//      This function initializes the MAPI libraries and logs into the
//      messaging subsystem.
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI InitMAPISession (BOOL fIsNTService, LPSTR pszProfileName)
{
    // Initialize the MAPI libraries before calling ANY MAPI function
    ULONG ulFlags = MAPI_NO_MAIL |
                    (pszProfileName ? MAPI_EXPLICIT_PROFILE : MAPI_USE_DEFAULT) |
                    MAPI_NEW_SESSION;
    
    MAPIINIT_0 init = { 0 };
    init.ulVersion  = MAPI_INIT_VERSION;
    init.ulFlags    = (fIsNTService ? MAPI_NT_SERVICE : 0);
    
    HRESULT hResult = MAPIInitialize (&init);
    if (hResult)
    {
        LogEvent (TRUE, EVENTMSG_E_FAILED_INIT_SESSION, hResult);
    }
    else
    {
        // Logon onto the message subsystem.
        hResult = MAPILogonEx (0,
                               pszProfileName,
                               NULL,
                               ulFlags,
                               &g_pSession);
        if (hResult)
        {
            LogEvent (TRUE, EVENTMSG_E_FAILED_MAPILOGONEX, hResult);
        }
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    FinishMAPISession()
//
//    Parameters
//      None
//      
//    Purpose
//      This functions logs off from the messaging subsystem. Here we clean up
//      all the MAPI objects. Also don't forget to uninitialize
//      the MAPI libraries.
//      
//    Return Value
//      None
//
void WINAPI FinishMAPISession()
{
    if (g_IdentityEID.lpb)
    {
        MAPIFreeBuffer (g_IdentityEID.lpb);
    }
    if (g_pFolder)
    {
        g_pFolder->Release();
    }
    if (g_pMDB)
    {
        ULONG ulFlags = LOGOFF_ORDERLY;
        g_pMDB->StoreLogoff (&ulFlags);
        g_pMDB->Release();
    }
    // If we have a session, terminate it
    if (g_pSession)
    {
        // Logoff from the messaging subsystem. Terminate and allow any UI
        // in case any provider or MAPI needs to display a UI.
        g_pSession->Logoff (0, MAPI_LOGOFF_UI, 0);
        g_pSession->Release();
        g_pSession = NULL;
    }
    // When we are exiting the app, uninitialize the MAPI libraries
    MAPIUninitialize();
}

// Exchange PR_MDB_PROVIDER GUID stolen from EDKMDB.H on the Exchange SDK
// NOTE: You should obtain the Exchange Developer's Kit and include
//       the file EDKMDB.H. We have included these definitions for
//       simplicity in the sample but this is not the ideal solution.
//              IrvingD 10/8/96
#define pbExchangeProviderPrimaryUserGuid	"\x54\x94\xA1\xC0\x29\x7F\x10\x1B\xA5\x87\x08\x00\x2B\x2A\x25\x17"
#define pbExchangeProviderDelegateGuid		"\x9e\xb4\x77\x00\x74\xe4\x11\xce\x8c\x5e\x00\xaa\x00\x42\x54\xe2"
#define pbExchangeProviderPublicGuid		"\x78\xb2\xfa\x70\xaf\xf7\x11\xcd\x9b\xc8\x00\xaa\x00\x2f\xc4\x5a"
#define pbExchangeProviderXportGuid			"\xa9\x06\x40\xe0\xd6\x93\x11\xcd\xaf\x95\x00\xaa\x00\x4a\x35\xc3"

#define	pidStoreMin						    0x6618
#define PR_IPM_PUBLIC_FOLDERS_ENTRYID		PROP_TAG (PT_BINARY, pidStoreMin+0x19)

///////////////////////////////////////////////////////////////////////////////
//    OpenExchangePFStore()
//
//    Parameters
//      None
//
//    Purpose
//      On the current MAPI session, open the message store table, search for
//      the Microsoft Exchange Public Folder provider and open it.
//      The UUID of the Exchange Public Folder provider is used to search
//      on the stores table.
//
//    Return Value
//      An HRESULT
//
HRESULT WINAPI OpenExchangePFStore()
{
    // Set an ENUM list for safe and sure index in the row's property
    enum
    {
        MST_EID,
        MST_NAME,
        MST_GUID,
        MSG_STORES_TABLE_PROPS  // Array Size
    };
    const static SizedSPropTagArray (MSG_STORES_TABLE_PROPS, sptMsgStores) =
    {
        MSG_STORES_TABLE_PROPS,
	    {
            PR_ENTRYID,
            PR_DISPLAY_NAME,
            PR_MDB_PROVIDER
        }
    };
    MAPIUID mdbUID = { 0 };
    CopyMemory (&mdbUID, pbExchangeProviderPublicGuid, sizeof(MAPIUID));
    SRestriction srExchPF;
    SPropValue spvExchPFMDB;
    spvExchPFMDB.ulPropTag     = PR_MDB_PROVIDER;
    spvExchPFMDB.Value.bin.cb  = sizeof(MAPIUID);
    spvExchPFMDB.Value.bin.lpb = (LPBYTE)&mdbUID;
    
    srExchPF.rt                        = RES_PROPERTY;
	srExchPF.res.resProperty.relop     = RELOP_EQ;
	srExchPF.res.resProperty.ulPropTag = spvExchPFMDB.ulPropTag;
	srExchPF.res.resProperty.lpProp    = &spvExchPFMDB;
    
    LPMAPITABLE pTable = NULL;
    LPSRowSet pRows = NULL;
    HRESULT hResult = g_pSession->GetMsgStoresTable (0, &pTable);
    if (hResult)
    {
        // Can't open the store tables? then we can't do anything else.
        LogEvent (TRUE, EVENTMSG_E_FAILED_GETMSGSTORETABLE, hResult);
    }
    else
    {
        // The message store tables is a relatively small table. For
        // efficiency we can query ALL rows at once.
        hResult = HrQueryAllRows (pTable,
                                  (LPSPropTagArray)&sptMsgStores,
                                  &srExchPF,
                                  NULL,
                                  0,
                                  &pRows);
        if (hResult || !pRows || pRows->cRows < 1)
        {
            LogEvent (TRUE, EVENTMSG_E_FAILED_HRQUERYALLROWS, hResult);
        }
        else    
        {
            hResult = g_pSession->OpenMsgStore (0,
                                                pRows->aRow->lpProps[MST_EID].Value.bin.cb,
                                                (LPENTRYID)pRows->aRow->lpProps[MST_EID].Value.bin.lpb,
                                                NULL,
                                                MDB_WRITE | MAPI_DEFERRED_ERRORS,
                                                &g_pMDB);
            if (hResult)
            {
                LogEvent (TRUE, EVENTMSG_E_FAILED_OPENMSGSTORE, hResult);
            }
            else
            {
                LPSPropValue pPFIPMEid;
                hResult = HrGetOneProp (g_pMDB, PR_IPM_PUBLIC_FOLDERS_ENTRYID, &pPFIPMEid);
                if (hResult)
                {
                    LogEvent (TRUE, EVENTMSG_E_FAILED_HRGETONEPROP, hResult);
                }
                else
                {
                    ULONG ulObjType;
                    LPMAPIFOLDER pPFIPMFld;
                    hResult = g_pMDB->OpenEntry (pPFIPMEid->Value.bin.cb,
                                                 (LPENTRYID)pPFIPMEid->Value.bin.lpb,
                                                 NULL,
                                                 MAPI_MODIFY | MAPI_DEFERRED_ERRORS,
                                                 &ulObjType,
                                                 (LPUNKNOWN *)&pPFIPMFld);
                    if (hResult)
                    {
                        LogEvent (TRUE, EVENTMSG_E_FAILED_OPENENTRY, hResult);
                    }
                    else
                    {
                        hResult = pPFIPMFld->CreateFolder (FOLDER_GENERIC,
                                                           "DIRMON Sample Posts",
                                                           NULL,
                                                           NULL,
                                                           MAPI_DEFERRED_ERRORS | OPEN_IF_EXISTS,
                                                           &g_pFolder);
                        if (hResult)
                        {
                            LogEvent (TRUE, EVENTMSG_E_FAILED_CREATEFOLDER, hResult);
                        }
                        pPFIPMFld->Release();
                    }
                    MAPIFreeBuffer (pPFIPMEid);
                }
            }
        }
        if (pRows)
        {
            FreeProws (pRows);
        }
        pTable->Release();
    }
    return hResult;
}

///////////////////////////////////////////////////////////////////////////////
//    TerminateService()
//
//    Parameters
//      [IN]    dwError         Error code with which the service will end
//      [IN]    fRunAsService   TRUE is we are running as a service.
//
//    Purpose
//      Close down all the allocated resources in preparation for the
//      service shutdown.
//
//    Return Value
//      A Win32 error code
//
DWORD WINAPI TerminateService (DWORD dwError, BOOL fRunAsService)
{
    if (g_hTerminateEvent)
    {
        CloseHandle (g_hTerminateEvent);
    }
    if (g_ssHandle)
    {
        SendStatusToSCM (SERVICE_STOPPED, dwError, 0, 0);
    }
    if (fRunAsService)
    {
        FinishMAPISession();
    }
    UnInitEventLog();
    return NO_ERROR;
}

// End of file for MAIN.CPP
